from functools import partial
import numpy as np
import numexpr as ne
from ducc0.wgridder.experimental import vis2dirty, dirty2vis
from ducc0.fft import r2c, c2r
from ducc0.misc import empty_noncritical
from pfb.operators.psf import (psf_convolve_slice,
                               psf_convolve_cube)
from pfb.opt.pcg import pcg
from pfb.utils.misc import taperf
from time import time

def _hessian_slice(x,
                   xout=None,
                   uvw=None,
                   weight=None,
                   vis_mask=None,
                   freq=None,
                   beam=None,
                   cell=None,
                   x0=0.0,
                   y0=0.0,
                   flip_u=False,
                   flip_v=True,
                   flip_w=False,
                   do_wgridding=True,
                   epsilon=1e-7,
                   double_accum=True,
                   nthreads=1,
                   eta=None,
                   wsum=None):
    '''
    Apply vis space Hessian approximation on a slice of an image.

    Important!
    x0, y0, flip_u, flip_v and flip_w must be consistent with the
    conventions defined in pfb.operators.gridder.wgridder_conventions

    These are inputs here to allow for testing but should generally be taken
    from the attrs of the datasets produced by
    pfb.operators.gridder.image_data_products
    '''
    if not x.any():
        return np.zeros_like(x)
    nx, ny = x.shape
    mvis = dirty2vis(uvw=uvw,
                    freq=freq,
                    mask=vis_mask,
                    dirty=x if beam is None else x * beam,
                    pixsize_x=cell,
                    pixsize_y=cell,
                    center_x=x0,
                    center_y=y0,
                    flip_u=flip_u,
                    flip_v=flip_v,
                    flip_w=flip_w,
                    epsilon=epsilon,
                    nthreads=nthreads,
                    do_wgridding=do_wgridding,
                    divide_by_n=False)

    convim = vis2dirty(
                    uvw=uvw,
                    freq=freq,
                    vis=mvis,
                    wgt=weight,
                    mask=vis_mask,
                    dirty=xout,  # return in case xout is None
                    npix_x=nx,
                    npix_y=ny,
                    pixsize_x=cell,
                    pixsize_y=cell,
                    center_x=x0,
                    center_y=y0,
                    flip_u=flip_u,
                    flip_v=flip_v,
                    flip_w=flip_w,
                    epsilon=epsilon,
                    nthreads=nthreads,
                    do_wgridding=do_wgridding,
                    double_precision_accumulation=double_accum,
                    divide_by_n=False)

    if wsum is not None:
        convim /= wsum

    if beam is not None:
        convim *= beam

    if eta is not None:
        convim += eta * x

    return convim


def _hessian_psf_slice(x,       # input image, not overwritten
                    xpad=None,  # preallocated array to store padded image
                    xhat=None,  # preallocated array to store FTd image
                    xout=None,  # preallocated array to store output image
                    abspsf=None,
                    beam=None,
                    lastsize=None,
                    nthreads=1,
                    eta=None):
    """
    Tikhonov regularised Hessian approx
    """
    nx, ny = x.shape
    xpad.fill(0.0)
    if beam is None:
        np.copyto(xpad[0:nx, 0:ny], x)
    else:
        xpad[0:nx, 0:ny] = x*beam
    r2c(xpad, axes=(0, 1), nthreads=nthreads,
        forward=True, inorm=0, out=xhat)
    xhat *= abspsf
    c2r(xhat, axes=(0, 1), forward=False, out=xpad,
        lastsize=lastsize, inorm=2, nthreads=nthreads,
        allow_overwriting_input=True)
    np.copyto(xout, xpad[0:nx, 0:ny])

    if beam is not None:
        xout *= beam

    if eta:
        xout += x * eta

    return xout


def hessian_psf_cube(
                    xpad,  # preallocated array to store padded image
                    xhat,  # preallocated array to store FTd image
                    xout,  # preallocated array to store output image
                    beam,
                    abspsf,
                    lastsize,
                    x,     # input image, not overwritten
                    nthreads=1,
                    eta=1,
                    mode='forward'):
    """
    Tikhonov regularised Hessian approx
    """
    if mode=='forward':
        if beam is not None:
            psf_convolve_cube(xpad, xhat, xout, abspsf, lastsize, x*beam,
                            nthreads=nthreads)
        else:
            psf_convolve_cube(xpad, xhat, xout, abspsf, lastsize, x,
                            nthreads=nthreads)

        if beam is not None:
            xout *= beam

        if eta:
            xout += x * eta

        return xout
    else:
        raise NotImplementedError


def hess_direct(x,     # input image, not overwritten
                xpad=None,  # preallocated array to store padded image
                xhat=None,  # preallocated array to store FTd image
                xout=None,  # preallocated array to store output image
                abspsf=None,
                taperxy=None,
                lastsize=None,
                nthreads=1,
                eta=1,
                mode='forward'):
    nband, nx, ny = x.shape
    xpad.fill(0.0)
    xpad[:, 0:nx, 0:ny] = x * taperxy[None]
    r2c(xpad, out=xhat, axes=(1,2),
        forward=True, inorm=0, nthreads=nthreads)
    if mode=='forward':
        xhat *= (abspsf + eta)
    else:
        xhat /= (abspsf + eta)
    c2r(xhat, axes=(1, 2), forward=False, out=xpad,
        lastsize=lastsize, inorm=2, nthreads=nthreads,
        allow_overwriting_input=True)
    np.copyto(xout, xpad[:, 0:nx, 0:ny])
    xout *= taperxy[None]
    return xout


def hess_direct_slice(x,     # input image, not overwritten
                xpad=None,  # preallocated array to store padded image
                xhat=None,  # preallocated array to store FTd image
                xout=None,  # preallocated array to store output image
                abspsf=None,
                taperxy=None,
                lastsize=None,
                nthreads=1,
                eta=1,
                mode='forward'):
    '''
    Note eta must be relative to wsum (peak of PSF)
    '''
    nx, ny = x.shape
    xpad.fill(0.0)
    xpad[0:nx, 0:ny] = x * taperxy
    r2c(xpad, out=xhat, axes=(0,1),
        forward=True, inorm=0, nthreads=nthreads)
    if mode=='forward':
        xhat *= (abspsf + eta)
    else:
        xhat /= (abspsf + eta)
    c2r(xhat, axes=(0, 1), forward=False, out=xpad,
        lastsize=lastsize, inorm=2, nthreads=nthreads,
        allow_overwriting_input=True)
    np.copyto(xout, xpad[0:nx, 0:ny])
    xout *= taperxy
    return xout


class hess_psf(object):
    def __init__(self, nx, ny, abspsf,
                 beam=None,
                 eta=1.0,
                 nthreads=1,
                 cgtol=1e-3,
                 cgmaxit=300,
                 cgverbose=2,
                 cgrf=25,
                 taper_width=32,
                 min_beam=5e-3):
        self.nx = nx
        self.ny = ny
        self.abspsf = abspsf
        self.nband, self.nx_psf, self.nyo2 = abspsf.shape
        if beam is not None and not (beam==1).all():
            assert self.nband == beam.shape[0]
            assert self.nx == beam.shape[1]
            assert self.ny == beam.shape[2]
            self.beam = beam
        else:
            # self.beam = None
            self.beam = (None,)*self.nband
        self.ny_psf = 2*(self.nyo2-1)
        self.nx_pad = self.nx_psf - self.nx
        self.ny_pad = self.ny_psf - self.ny
        self.nthreads = nthreads
        if isinstance(eta, float):
            self.eta = np.tile(eta, self.nband)
        else:
            try:
                self.eta = np.array(eta)
                assert self.eta.size == self.nband
            except Exception as e:
                raise e


        # per band tmp arrays
        self.xhat = empty_noncritical((self.nx_psf, self.nyo2),
                                      dtype='c16')
        self.xpad = empty_noncritical((self.nx_psf, self.ny_psf),
                                      dtype='f8')
        self.xout = empty_noncritical((self.nband, self.nx, self.ny),
                                      dtype='f8')

        # conjugate gradient params
        self.cgtol=cgtol
        self.cgmaxit=cgmaxit
        self.cgverbose=cgverbose
        self.cgrf=cgrf

        # taper for direct mode
        self.taperxy = taperf((nx, ny), taper_width)

        # for beam application in direct mode
        self.min_beam = min_beam

    def set_beam(self, beam):
        assert beam.shape == (self.nband, self.nx, self.ny)
        self.beam = beam

    def dot(self, x):
        if len(x.shape) == 3:
            xtmp = x
        elif len(x.shape) == 2:
            xtmp = x[None, :, :]
        else:
            raise ValueError("Unsupported number of input dimensions")

        nband, nx, ny = xtmp.shape
        assert nband == self.nband
        assert nx == self.nx
        assert ny == self.ny

        for b in range(nband):
            self.xpad.fill(0.0)
            if self.beam[b] is None:
                np.copyto(self.xpad[0:nx, 0:ny], xtmp[b])
            else:
                self.xpad[0:nx, 0:ny] = xtmp[b]*self.beam[b]
            r2c(self.xpad, axes=(0, 1), nthreads=self.nthreads,
                forward=True, inorm=0, out=self.xhat)
            self.xhat *= self.abspsf[b]
            c2r(self.xhat, axes=(0, 1), forward=False, out=self.xpad,
                lastsize=self.ny_psf, inorm=2, nthreads=self.nthreads,
                allow_overwriting_input=True)
            if self.beam[b] is None:
                np.copyto(self.xout[b], self.xpad[0:nx, 0:ny])
            else:
                self.xout[b] = self.xpad[0:nx, 0:ny]*self.beam[b]
        self.xout += xtmp * self.eta[:, None, None]
        return self.xout

    def hdot(self, x):
        # Hermitian operator
        return self.dot(x)

    def idot(self, x, mode='psf', x0=None, init_x0=True):
        if len(x.shape) == 3:
            xtmp = x
        elif len(x.shape) == 2:
            xtmp = x[None, :, :]
        else:
            raise ValueError("Unsupported number of dimensions")

        nband, nx, ny = xtmp.shape
        assert nband == self.nband
        assert nx == self.nx
        assert ny == self.ny

        if x0 is None and init_x0:
            # initialise with direct estimate
            x0 = np.zeros_like(xtmp)
            for b in range(self.nband):
                x0[b] = hess_direct_slice(xtmp[b],
                                    xpad=self.xpad,
                                    xhat=self.xhat,
                                    xout=self.xout[b],
                                    abspsf=self.abspsf[b],
                                    taperxy=self.taperxy,
                                    lastsize=self.ny_psf,
                                    nthreads=self.nthreads,
                                    eta=self.eta[b]*np.sqrt(nx*ny),
                                    mode='backward')
                if self.beam[b] is not None:
                    mask = (self.xout[b] > 0) & (self.beam[b] > self.min_beam)
                    self.xout[b, mask] /= self.beam[b, mask]**2
        else:
            x0 = np.zeros_like(xtmp)

        if mode=='direct':
            for b in range(self.nband):
                self.xout[b] = hess_direct_slice(xtmp[b],
                                    xpad=self.xpad,
                                    xhat=self.xhat,
                                    xout=self.xout[b],
                                    abspsf=self.abspsf[b],
                                    taperxy=self.taperxy,
                                    lastsize=self.ny_psf,
                                    nthreads=self.nthreads,
                                    eta=self.eta[b]*np.sqrt(nx*ny),
                                    mode='backward')
                if self.beam[b] is not None:
                    mask = (self.xout[b] > 0) & (self.beam[b] > self.min_beam)
                    self.xout[b, mask] /= self.beam[b, mask]**2

        elif mode=='psf':
            for b in range(self.nband):
                hess = partial(_hessian_psf_slice,
                            xpad=self.xpad,
                            xhat=self.xhat,
                            xout=self.xout[b],
                            abspsf=self.abspsf[b],
                            beam=self.beam[b],
                            lastsize=self.ny_psf,
                            nthreads=self.nthreads,
                            eta=self.eta[b])
                self.xout[b] = pcg(hess,
                            xtmp[b],
                            x0=x0[b],
                            tol=self.cgtol,
                            maxit=self.cgmaxit,
                            minit=3,
                            verbosity=self.cgverbose,
                            report_freq=self.cgrf,
                            backtrack=False,
                            return_resid=False)
        else:
            raise ValueError(f"Unknown mode {mode}")

        return self.xout.copy()


##################### jax #####################################
import jax
import jax.numpy as jnp
from functools import partial

@partial(jax.jit, static_argnums=(0,1,2,3,4))
def hessian_slice_jax(
                    nx, ny,
                    nx_psf, ny_psf,
                    eta,
                    psfhat,
                    x):
    psfh = jax.lax.stop_gradient(psfhat)
    xhat = jnp.fft.rfft2(x,
                         s=(nx_psf, ny_psf),
                         norm='backward')
    xout = jnp.fft.irfft2(xhat*psfh,
                          s=(nx_psf, ny_psf),
                          norm='backward')[0:nx, 0:ny]
    return xout + eta*x

@partial(jax.jit, static_argnums=(0,1,2,3,4))
def hessian_jax(nx, ny,
                nx_psf, ny_psf,
                eta,
                psfhat,
                x):
    psfh = jax.lax.stop_gradient(psfhat)
    xhat = jnp.fft.rfft2(x,
                         s=(nx_psf, ny_psf),
                         norm='backward')
    xout = jnp.fft.irfft2(xhat*psfh,
                          s=(nx_psf, ny_psf),
                          norm='backward')[:, 0:nx, 0:ny]
    return xout + eta*x

@partial(jax.jit, static_argnums=(0,1,2,3,4))
def hessian_jax(nx, ny,
                nx_psf, ny_psf,
                eta,
                psfhat,
                x):
    psfh = jax.lax.stop_gradient(psfhat)
    xhat = jnp.fft.rfft2(x,
                         s=(nx_psf, ny_psf),
                         norm='backward')
    xout = jnp.fft.irfft2(xhat*psfh,
                          s=(nx_psf, ny_psf),
                          norm='backward')[:, 0:nx, 0:ny]
    return xout + eta*x

@partial(jax.jit, static_argnums=(0,1,2,3,4))
def fshessian_jax(nx, ny,
                  nx_psf, ny_psf,
                  eta,
                  mask,
                  psfhat,
                  x):
    '''
    Assumes both x and psfhat is a 4D cube with the last two dimensions
    corresponding to spatial dimensions along which FFT us taken.

    nx, ny          - int: npix in two spatial coordinates
    nx_psf, ny_psf  - int: npix in two spatial coordinates of psf 
    eta             - float: regularisation parameter
    mask            - ndarray(nx, ny): spatial mask
    psfhat          - ndarray(nband, ncorr, nx, ny): psf in uv space
    x               - ndarray(nband, ncorr, nx, ny): image
    '''
    mask = jax.lax.stop_gradient(mask)[None, None, :, :]
    psfhat = jax.lax.stop_gradient(psfhat)
    xhat = jnp.fft.rfft2(x*mask,
                         s=(nx_psf, ny_psf),
                         norm='backward')
    xout = jnp.fft.irfft2(xhat*psfhat,
                          s=(nx_psf, ny_psf),
                          norm='backward')[:, :, 0:nx, 0:ny]
    return xout*mask + eta*x


import numpy as np
from scipy.interpolate import RegularGridInterpolator as RGI
from katbeam import JimBeam
from africanus.rime.fast_beam_cubes import beam_cube_dde
from africanus.rime import parallactic_angles, feed_rotation
from pfb.utils.stokes import jones_to_mueller, mueller_to_stokes
from scipy.ndimage import rotate
from pfb.utils.fits import set_wcs
from astropy.wcs import WCS
from reproject import reproject_interp

def interp_beam(freq, nx, ny, cell_deg, btype,
                utime=None, ant_pos=None, phase_dir=None):
    '''
    A function that returns an object array containing a function
    returning beam values given (l,m) coordinates at a single frequency.
    Frequency mapped to imaging band extenally. Result is meant to be
    passed into eval_beam below.
    '''
    if isinstance(freq, np.ndarray):
        assert freq.size == 1, "Only single frequency interpolation currently supported"
        freq = freq[0]
    if btype is None:
        l = (-(nx//2) + np.arange(nx)) * cell_deg
        m = (-(ny//2) + np.arange(ny)) * cell_deg
        return np.ones((nx, ny), dtype=float), l, m
    elif btype.endswith('.npz'):
        # these are expected to be in the format given here
        # https://archive-gw-1.kat.ac.za/public/repository/10.48479/wdb0-h061/index.html
        dct = np.load(btype)
        beam = dct['abeam']
        l = np.deg2rad(dct['ldeg'])
        m = np.deg2rad(dct['mdeg'])
        ll, mm = np.meshgrid(l, m, indexing='ij')
        bfreqs = dct['freq']
        beam_amp = (beam[0, :, :, :] * beam[0, :, :, :].conj() +
                    beam[-1, :, :, :] * beam[-1, :, :, :].conj())/2.0
        beam_amp = np.transpose(beam_amp, (1,2,0))[:, :, :, None, None].real
    else:
        btype = btype.lower()
        btype = btype.replace('-', '_')
        l = (-(nx//2) + np.arange(nx)) * cell_deg
        m = (-(ny//2) + np.arange(ny)) * cell_deg
        ll, mm = np.meshgrid(l, m, indexing='ij')
        if btype in ["kbl", "kb_l", "katbeam_l"]:
            # katbeam L band
            beam_amp = JimBeam('MKAT-AA-L-JIM-2020').I(ll.flatten(),
                                                       mm.flatten(),
                                                       freqMHz=freq/1e6)
        elif btype in ["kbuhf", "kb_uhf", "katbeam_uhf"]:
            # katbeam L band
            beam_amp = JimBeam('MKAT-AA-UHF-JIM-2020').I(ll.flatten(),
                                                       mm.flatten(),
                                                       freqMHz=freq/1e6)
        else:
            raise ValueError(f"Unknown beam model {btype}")
        beam_amp = beam_amp.reshape(nx, ny)[:, :, None, None, None]
        bfreqs = np.array((freq,))

    if utime is None:
        return beam_amp.squeeze(), l, m

    parangles = parallactic_angles(utime, ant_pos, phase_dir, backend='astropy')
    # mean over antanna nant -> 1
    parangles = np.mean(parangles, axis=1, keepdims=True)
    nant = 1
    # beam_cube_dde requirements
    nband = 1
    ntimes = utime.size
    ant_scale = np.ones((nant, nband, 2), dtype=np.float64)
    point_errs = np.zeros((ntimes, nant, nband, 2), dtype=np.float64)
    beam_extents = np.array([[l.min(), l.max()], [m.min(), m.max()]])
    lm = np.vstack((ll.flatten(), mm.flatten())).T
    beam_image = beam_cube_dde(np.ascontiguousarray(beam_amp),
                            beam_extents, bfreqs,
                            lm, parangles, point_errs,
                            ant_scale, np.array((freq,))).squeeze()

    return beam_image.squeeze(), l, m


def eval_beam(beam_image, l_in, m_in, l_out, m_out):
    if l_out.ndim == 2:
        ll = l_out
        mm = m_out
    elif l_out.ndim == 1:
        ll, mm = np.meshgrid(l_out, m_out, indexing='ij')
    else:
        msg = 'Only 1 or 2D coordinates supported for beam evaluation'
        raise ValueError(msg)

    if (beam_image == 1.0).all():
        return np.ones_like(ll)
    else:  # this gets expensive
        beamo = RGI((l_in, m_in), beam_image,
                    bounds_error=False, method='linear', fill_value=1.0)
        return beamo((ll, mm))

def reproject_and_interp_beam(beam, time, antpos, radec0, radecf,
                              cell_deg_in, cell_deg_out, nxo, nyo,
                              poltype, product, weight=None, nthreads=1):
    '''
    beam    - (2, 2, nx, ny)
    time    - nrow
    antpos  - (nant, 3)
    radec0  - original pointing direction
    radecf  - direction to project to
    '''
    utime = np.unique(time)
    ntime = utime.size
    # parangles = parallactic_angles(utime, antpos, np.array(radec0))
    # # use the mean over antenna
    # parangles = np.mean(parangles, axis=-1, keepdims=False)
    _, _, nxi, nyi = beam.shape
    # beamo = np.zeros((ntime, 2, 2, nxi, nyi), dtype=beam.dtype)
    # for i, parang in enumerate(parangles):
    #     # spatial rotation (assuming position angle is paralactic angle i.e. linear North-South, East-West interferometer)
    #     beamo[i] = rotate(beam, parang, axes=(-2, -1), reshape=False, order=1, mode='nearest')
    #     # feed rotation
    #     beamo[i] = rotate(beamo[i], parang, axes=(0, 1), reshape=False, order=1, mode='nearest')

    # # compute the weighted sum over time
    # if weight is not None and ntime > 1:
    #     wsumt = np.zeros((ntime, 2, 2))
    #     for i, t in enumerate(utime):
    #         sel = time==t
    #         wsumt[i] = weight[sel].sum(axis=(0, 1)).reshape(2, 2)
    #     beamo = np.sum(wsumt[:, :, :, None, None] * beamo, axis=0)
    #     beamo /= np.sum(wsumt, axis=0)[:, :, None, None]
    # else:
    #     beamo = np.mean(beamo, axis=0)

    # jones to Mueller
    beamo = beam
    beamo = jones_to_mueller(beamo, beamo)

    # Mueller to Stokes
    beamo = mueller_to_stokes(beamo, poltype=poltype)
    
    # select required products
    i = ()
    if 'I' in product:
        i += (0,)
    if 'Q' in product:
        i += (1,)
    if 'U' in product:
        i += (2,)
    if 'V' in product:
        i += (3,)
    beamo = beamo[i, ...]
    
    # reproject onto target field
    wcs_ref = WCS(naxis=2)
    wcs_ref.wcs.ctype = ['RA---SIN', 'DEC--SIN']
    wcs_ref.wcs.cdelt = np.array((cell_deg_in, cell_deg_in))
    wcs_ref.wcs.cunit = ['deg', 'deg']
    wcs_ref.wcs.crval = np.array((radec0[0]*180.0/np.pi, radec0[1]*180.0/np.pi))
    wcs_ref.wcs.crpix = [1 + nxi//2, 1 + nyi//2]
    wcs_ref.array_shape = [nxi, nyi]

    # header for target field
    wcs_target = WCS(naxis=2)
    wcs_target.wcs.ctype = ['RA---SIN', 'DEC--SIN']
    wcs_target.wcs.cdelt = np.array((cell_deg_out, cell_deg_out))
    wcs_target.wcs.cunit = ['deg', 'deg']
    wcs_target.wcs.crval = np.array((radecf[0]*180.0/np.pi, radecf[1]*180.0/np.pi))
    wcs_target.wcs.crpix = [nxo//2, nyo//2]
    wcs_target.array_shape = [nxo, nyo]

    pbeam = np.zeros((len(product), nxo, nyo), dtype=beamo.dtype)
    pmask = np.zeros((len(product), nxo, nyo), dtype=beamo.dtype)
    for i in range(len(product)):
        pbeam[i], pmask[i] = reproject_interp((beamo[i], wcs_ref), wcs_target, shape_out=(nxo, nyo))  # , block_size='auto', parallel=nthreads

    # set beam to zero where it is not defined
    pmask = pmask.astype(bool)
    pbeam[~pmask] = 0.0
    return pbeam


import fsspec
import s3fs
from functools import partial
import pickle
import xarray as xr
import concurrent.futures as cf

def set_output_names(opts):
    '''
    Make sure base folders exist and set default naming conventions
    '''
    output_filename = opts.output_filename
    product = opts.product
    fits_output_folder = opts.fits_output_folder
    log_directory = opts.log_directory

    if '://' in output_filename:
        protocol = output_filename.split('://')[0]
        prefix = f'{protocol}://'
    else:
        protocol = 'file'
        prefix = ''

    fs = fsspec.filesystem(protocol)
    basedir = fs.expand_path('/'.join(output_filename.split('/')[:-1]))[0]
    if not fs.exists(basedir):
        fs.makedirs(basedir)

    oname = output_filename.split('/')[-1] + f'_{product.upper()}'

    opts.output_filename = f'{prefix}{basedir}/{oname}'

    if fits_output_folder is not None:
        # this should be a file system
        fs = fsspec.filesystem('file')
        fbasedir = fs.expand_path(fits_output_folder)[0]
        if not fs.exists(fbasedir):
            fs.makedirs(fbasedir)
        fits_output_folder = fbasedir
    else:
        if protocol != 'file':
            raise ValueError('You must provide a separate fits-output-'
                             'folder when output protocol is '
                             f'{protocol}')
        fits_output_folder = basedir

    opts.fits_output_folder = fits_output_folder

    if log_directory is not None:
        # this should be a file system
        fs = fsspec.filesystem('file')
        lbasedir = fs.expand_path(log_directory)[0]
        if not fs.exists(lbasedir):
            fs.makedirs(lbasedir)
        log_directory = lbasedir
    else:
        if protocol != 'file':
            raise ValueError('You must provide a separate log-'
                             'directory when output protocol is '
                             f'{protocol}')
        log_directory = basedir

    opts.log_directory = log_directory

    return opts, basedir, oname


def xds_from_url(url, columns='ALL', chunks=-1):
    '''
    Returns a lazy view of all datasets contained in url
    as well as a list of full paths to each one
    '''
    if columns.upper() != 'ALL':
        raise NotImplementedError
    if chunks != -1:
        raise NotImplementedError

    url = url.rstrip('/')
    if '://' in url:
        protocol = url.split('://')[0]
        prefix = f'{protocol}://'
    else:
        protocol = 'file'
        prefix = ''
    fs = fsspec.filesystem(protocol)
    ds_list = fs.glob(f'{url}/*.zarr')
    ds_list = list(map(lambda x: prefix + x, ds_list))
    # these will only be read in on first value access and won't be chunked
    open_zarr = partial(xr.open_zarr, chunks=None)
    xds = list(map(open_zarr, ds_list))

    if not len(xds):
        raise ValueError(f'Nothing found at {url}')
    return xds, ds_list


def read_var(ds, var):
    '''
    Access var to force loading it into memory
    '''
    val = getattr(ds, var).values
    return 1


def xds_from_list(ds_list, drop_vars=None, drop_all_but=None,
                  chunks=-1, nthreads=1, order_freq=True):
    '''
    Reads a list of datasets into memory in parallel.
    Use drop_vars to drop vars that should not be read into memory.
    Shoudl return a list that is ordered in frequency
    '''
    if chunks != -1:
        raise NotImplementedError

    # these will only be read in on first value access and won't be chunked
    open_zarr = partial(xr.open_zarr, chunks=None)
    xds = list(map(open_zarr, ds_list))
    if not len(xds):
        raise ValueError(f'Nothing found at {ds_list[0]}')

    if drop_vars is not None:
        if not isinstance(drop_vars, list):
            drop_vars = list(drop_vars)
        for i, ds in enumerate(xds):
            xds[i] = ds.drop_vars(drop_vars, errors="ignore")

    if drop_all_but is not None:
        if not isinstance(drop_all_but, list):
            drop_all_but = [drop_all_but]
        for i, ds in enumerate(xds):
            drop_vars = [var for var in ds.data_vars]
            for var in drop_all_but:
                if var in drop_vars:
                    drop_vars.pop(drop_vars.index(var))
            xds[i] = ds.drop_vars(drop_vars, errors="ignore")

    futures = []
    with cf.ThreadPoolExecutor(max_workers=nthreads) as executor:
        for ds in xds:
            for var in ds.data_vars:
                futures.append(executor.submit(read_var, ds, var))
    cf.wait(futures)

    # make sure list is ordered correctly (increasing frequency)
    # TODO - reorder if not
    if len(xds) > 1 and order_freq:
        freq_out = [ds.freq_out for ds in xds]
        freq_min = freq_out[0]
        for f in freq_out[1:]:
            if freq_min > f:
                raise ValueError('xds list is not ordered in frequency')
            freq_min = f
    return xds


def cache_opts(opts, url, protocol, name='opts.pkl'):
    fs = fsspec.filesystem(protocol)
    if protocol == 'file':
        with fs.open(f'{url}/{name}', 'wb') as f:
            pickle.dump(opts, f)
    elif protocol == 's3':
        s3_client = s3fs.S3FileSystem(anon=False)
        bucket = url.split('://')[1]
        s3_path = f"{bucket}/{name}"
        with s3_client.open(s3_path, 'wb') as f:
            pickle.dump(opts, f)
    else:
        raise ValueError(f"protocol {protocol} not yet supported")

def get_opts(url, protocol, name='opts.pkl'):
    fs = fsspec.filesystem(protocol)
    if protocol == 'file':
        with fs.open(f'{url}/{name}', 'rb') as f:
            optsp = pickle.load(f)
    elif protocol == 's3':
        s3_client = s3fs.S3FileSystem(anon=False)
        bucket = url.split('://')[1]
        s3_path = f"{bucket}/{name}"
        with s3_client.open(s3_path, 'rb') as f:
            optsp = pickle.load(f)
    else:
        raise ValueError(f"protocol {protocol} not yet supported")

    return optsp
